"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from . import utils
from .application import ApplicationInstallParams
from .asset import Asset
from .connections import PartialConnection
from .enums import PremiumType, try_enum
from .flags import ApplicationFlags
from .member import Member
from .mixins import Hashable
from .user import User

if TYPE_CHECKING:
    from datetime import datetime

    from .guild import Guild
    from .state import ConnectionState
    from .types.profile import (
        Profile as ProfilePayload,
        ProfileApplication as ProfileApplicationPayload,
        MutualGuild as MutualGuildPayload,
    )
    from .types.user import PartialUser as PartialUserPayload

__all__ = (
    'ApplicationProfile',
    'MutualGuild',
    'UserProfile',
    'MemberProfile',
)


class Profile:
    if TYPE_CHECKING:
        id: int
        bot: bool
        _state: ConnectionState

    def __init__(self, **kwargs) -> None:
        data: ProfilePayload = kwargs.pop('data')
        user = data['user']
        mutual_friends: List[PartialUserPayload] = kwargs.pop('mutual_friends', None)

        member = data.get('guild_member')
        if member is not None:
            member['user'] = user
            kwargs['data'] = member
        else:
            kwargs['data'] = user

        # n.b. this class is subclassed by UserProfile and MemberProfile
        # which subclass either User or Member respectively
        # Because of this, we call super().__init__ here
        # after ensuring the data kwarg is set correctly
        super().__init__(**kwargs)
        state = self._state

        self.bio: Optional[str] = user['bio'] or None

        # We need to do a bit of a hack here because premium_since is massively overloaded
        guild_premium_since = getattr(self, 'premium_since', utils.MISSING)
        if guild_premium_since is not utils.MISSING:
            self.guild_premium_since = guild_premium_since

        self.premium_type: Optional[PremiumType] = (
            try_enum(PremiumType, data['premium_type']) if user.get('premium_type') else None
        )
        self.premium_since: Optional[datetime] = utils.parse_time(data['premium_since'])
        self.premium_guild_since: Optional[datetime] = utils.parse_time(data['premium_guild_since'])
        self.connections: List[PartialConnection] = [PartialConnection(d) for d in data['connected_accounts']]

        self.mutual_guilds: Optional[List[MutualGuild]] = (
            [MutualGuild(state=state, data=d) for d in data['mutual_guilds']] if 'mutual_guilds' in data else None
        )
        self.mutual_friends: Optional[List[User]] = self._parse_mutual_friends(mutual_friends)
        self._mutual_friends_count: Optional[int] = data.get('mutual_friends_count')

        application = data.get('application')
        self.application: Optional[ApplicationProfile] = ApplicationProfile(data=application) if application else None

    def _parse_mutual_friends(self, mutual_friends: List[PartialUserPayload]) -> Optional[List[User]]:
        if self.bot:
            # Bots don't have friends
            return []
        if mutual_friends is None:
            return

        state = self._state
        return [state.store_user(friend) for friend in mutual_friends]

    @property
    def mutual_friends_count(self) -> Optional[int]:
        """Optional[:class:`int`]: The number of mutual friends the user has with the client user."""
        if self.bot:
            # Bots don't have friends
            return 0
        if self._mutual_friends_count is not None:
            return self._mutual_friends_count
        if self.mutual_friends is not None:
            return len(self.mutual_friends)

    @property
    def premium(self) -> bool:
        """:class:`bool`: Indicates if the user is a premium user."""
        return self.premium_since is not None


class ApplicationProfile(Hashable):
    """Represents a Discord application profile.

    .. container:: operations

        .. describe:: x == y

            Checks if two applications are equal.

        .. describe:: x != y

            Checks if two applications are not equal.

        .. describe:: hash(x)

            Return the applications's hash.

    .. versionadded:: 2.0

    Attributes
    ------------
    id: :class:`int`
        The application's ID.
    verified: :class:`bool`
        Indicates if the application is verified.
    popular_application_command_ids: List[:class:`int`]
        A list of the IDs of the application's popular commands.
    primary_sku_id: Optional[:class:`int`]
        The application's primary SKU ID, if any.
        This can be an application's game SKU, subscription SKU, etc.
    custom_install_url: Optional[:class:`str`]
        The custom URL to use for authorizing the application, if specified.
    install_params: Optional[:class:`ApplicationInstallParams`]
        The parameters to use for authorizing the application, if specified.
    """

    __slots__ = (
        'id',
        'verified',
        'popular_application_command_ids',
        'primary_sku_id',
        '_flags',
        'custom_install_url',
        'install_params',
    )

    def __init__(self, data: ProfileApplicationPayload) -> None:
        self.id: int = int(data['id'])
        self.verified: bool = data.get('verified', False)
        self.popular_application_command_ids: List[int] = [int(id) for id in data.get('popular_application_command_ids', [])]
        self.primary_sku_id: Optional[int] = utils._get_as_snowflake(data, 'primary_sku_id')
        self._flags: int = data.get('flags', 0)

        params = data.get('install_params')
        self.custom_install_url: Optional[str] = data.get('custom_install_url')
        self.install_params: Optional[ApplicationInstallParams] = (
            ApplicationInstallParams.from_application(self, params) if params else None
        )

    def __repr__(self) -> str:
        return f'<ApplicationProfile id={self.id} verified={self.verified}>'

    @property
    def flags(self) -> ApplicationFlags:
        """:class:`ApplicationFlags`: The flags of this application."""
        return ApplicationFlags._from_value(self._flags)

    @property
    def install_url(self) -> Optional[str]:
        """:class:`str`: The URL to install the application."""
        return self.custom_install_url or self.install_params.url if self.install_params else None

    @property
    def primary_sku_url(self) -> Optional[str]:
        """:class:`str`: The URL to the primary SKU of the application, if any."""
        if self.primary_sku_id:
            return f'https://discord.com/store/skus/{self.primary_sku_id}/unknown'


class MutualGuild(Hashable):
    """Represents a mutual guild between a user and the client user.

    .. container:: operations

        .. describe:: x == y

            Checks if two guilds are equal.

        .. describe:: x != y

            Checks if two guilds are not equal.

        .. describe:: hash(x)

            Returns the guild's hash.

    .. versionadded:: 2.0

    Attributes
    ------------
    id: :class:`int`
        The guild's ID.
    nick: Optional[:class:`str`]
        The guild specific nickname of the user.
    """

    __slots__ = ('id', 'nick', '_state')

    def __init__(self, *, state: ConnectionState, data: MutualGuildPayload) -> None:
        self._state = state
        self.id: int = int(data['id'])
        self.nick: Optional[str] = data.get('nick')

    def __repr__(self) -> str:
        return f'<MutualGuild guild={self.guild!r} nick={self.nick!r}>'

    @property
    def guild(self) -> Guild:
        """:class:`Guild`: The guild that the user is mutual with."""
        return self._state._get_or_create_unavailable_guild(self.id)


class UserProfile(Profile, User):
    """Represents a Discord user's profile.

    This is a :class:`User` with extended attributes.

    .. container:: operations

        .. describe:: x == y

            Checks if two users are equal.

        .. describe:: x != y

            Checks if two users are not equal.

        .. describe:: hash(x)

            Return the user's hash.

        .. describe:: str(x)

            Returns the user's name with discriminator.

    .. versionadded:: 2.0

    Attributes
    -----------
    application: Optional[:class:`ApplicationProfile`]
        The application profile of the user, if it is a bot.
    bio: Optional[:class:`str`]
        The user's "about me" field. Could be ``None``.
    premium_type: Optional[:class:`PremiumType`]
        Specifies the type of premium a user has (i.e. Nitro, Nitro Classic, or Nitro Basic). Could be None if the user is not premium.
    premium_since: Optional[:class:`datetime.datetime`]
        An aware datetime object that specifies how long a user has been premium (had Nitro).
        ``None`` if the user is not a premium user.
    premium_guild_since: Optional[:class:`datetime.datetime`]
        An aware datetime object that specifies when a user first Nitro boosted a guild.
    connections: Optional[List[:class:`PartialConnection`]]
        The connected accounts that show up on the profile.
    mutual_guilds: Optional[List[:class:`MutualGuild`]]
        A list of guilds that you share with the user.
        ``None`` if you didn't fetch mutual guilds.
    mutual_friends: Optional[List[:class:`User`]]
        A list of friends that you share with the user.
        ``None`` if you didn't fetch mutual friends.
    """

    __slots__ = (
        'bio',
        'premium_type',
        'premium_since',
        'premium_guild_since',
        'connections',
        'mutual_guilds',
        'mutual_friends',
        '_mutual_friends_count',
        'application',
    )

    def __repr__(self) -> str:
        return f'<UserProfile id={self.id} name={self.name!r} discriminator={self.discriminator!r} bot={self.bot} system={self.system} premium={self.premium}>'

    @property
    def display_bio(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the user's display bio.

        This is the same as :attr:`bio` and is here for compatibility.
        """
        return self.bio


class MemberProfile(Profile, Member):
    """Represents a Discord member's profile.

    This is a :class:`Member` with extended attributes.

    .. container:: operations

        .. describe:: x == y

            Checks if two members are equal.
            Note that this works with :class:`User` instances too.

        .. describe:: x != y

            Checks if two members are not equal.
            Note that this works with :class:`User` instances too.

        .. describe:: hash(x)

            Returns the member's hash.

        .. describe:: str(x)

            Returns the member's name with the discriminator.

    .. versionadded:: 2.0

    Attributes
    -----------
    application: Optional[:class:`ApplicationProfile`]
        The application profile of the user, if it is a bot.
    bio: Optional[:class:`str`]
        The user's "about me" field. Could be ``None``.
    guild_bio: Optional[:class:`str`]
        The user's "about me" field for the guild. Could be ``None``.
    guild_premium_since: Optional[:class:`datetime.datetime`]
        An aware datetime object that specifies the date and time in UTC when the member used their
        "Nitro boost" on the guild, if available. This could be ``None``.

        .. note::

            This is renamed from :attr:`Member.premium_since` because of name collisions.
    premium_type: Optional[:class:`PremiumType`]
        Specifies the type of premium a user has (i.e. Nitro, Nitro Classic, or Nitro Basic). Could be ``None`` if the user is not premium.
    premium_since: Optional[:class:`datetime.datetime`]
        An aware datetime object that specifies how long a user has been premium (had Nitro).
        ``None`` if the user is not a premium user.

        .. note::

            This is not the same as :attr:`Member.premium_since`. That is renamed to :attr:`guild_premium_since`.
    premium_guild_since: Optional[:class:`datetime.datetime`]
        An aware datetime object that specifies when a user first Nitro boosted a guild.
    connections: Optional[List[:class:`PartialConnection`]]
        The connected accounts that show up on the profile.
    mutual_guilds: Optional[List[:class:`MutualGuild`]]
        A list of guilds that you share with the user.
        ``None`` if you didn't fetch mutuals.
    mutual_friends: Optional[List[:class:`User`]]
        A list of friends that you share with the user.
        ``None`` if you didn't fetch mutuals.
    """

    __slots__ = (
        'bio',
        'guild_premium_since',
        'premium_type',
        'premium_since',
        'premium_guild_since',
        'connections',
        'mutual_guilds',
        'mutual_friends',
        '_mutual_friends_count',
        'application',
        '_banner',
        'guild_bio',
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        data = kwargs['data']
        member = data['guild_member']
        self._banner: Optional[str] = member.get('banner')
        self.guild_bio: Optional[str] = member.get('bio') or None

    def __repr__(self) -> str:
        return (
            f'<MemberProfile id={self._user.id} name={self._user.name!r} discriminator={self._user.discriminator!r}'
            f' bot={self._user.bot} nick={self.nick!r} premium={self.premium} guild={self.guild!r}>'
        )

    @property
    def display_banner(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns the member's display banner.

        For regular members this is just their banner (if available), but
        if they have a guild specific banner then that
        is returned instead.
        """
        return self.guild_banner or self._user.banner

    @property
    def guild_banner(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns an :class:`Asset` for the guild banner
        the member has. If unavailable, ``None`` is returned.
        """
        if self._banner is None:
            return None
        return Asset._from_guild_banner(self._state, self.guild.id, self.id, self._banner)

    @property
    def display_bio(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns the member's display bio.

        For regular members this is just their bio (if available), but
        if they have a guild specific bio then that
        is returned instead.
        """
        return self.guild_bio or self.bio



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import List, Optional, TypedDict
from typing_extensions import NotRequired

from .snowflake import Snowflake
from .subscriptions import SubscriptionTrial


class Promotion(TypedDict):
    id: Snowflake
    trial_id: NotRequired[Snowflake]
    start_date: str
    end_date: str
    flags: int
    outbound_title: str
    outbound_redemption_modal_body: str
    outbound_redemption_page_link: NotRequired[str]
    outbound_redemption_url_format: NotRequired[str]
    outbound_restricted_countries: NotRequired[List[str]]
    outbound_terms_and_conditions: str
    inbound_title: NotRequired[str]
    inbound_body_text: NotRequired[str]
    inbound_help_center_link: NotRequired[str]
    inbound_restricted_countries: NotRequired[List[str]]


class ClaimedPromotion(TypedDict):
    promotion: Promotion
    code: str
    claimed_at: str


class TrialOffer(TypedDict):
    id: Snowflake
    expires_at: str
    trial_id: Snowflake
    subscription_trial: SubscriptionTrial


class PromotionalPrice(TypedDict):
    amount: int
    currency: str


class PricingPromotion(TypedDict):
    plan_id: Snowflake
    country_code: str
    payment_source_types: List[str]
    price: PromotionalPrice


class WrappedPricingPromotion(TypedDict):
    country_code: str
    localized_pricing_promo: Optional[PricingPromotion]



import json
import time
import random
import base64
import requests
import websocket
from requests.structures import CaseInsensitiveDict
from threading import Thread
from .utils.payloads import StartSocket
from .utils.objects import MessageContent, AuthorProfile
import os


class SocketDiscord:
    def __init__(self, client):
        self.events = {}
        self.ws = None
        self.token = client.token
        self.client = client
        
        
        self.session_id = None
        

    def on(self, event):
        def event_ready(handler):
            if event not in self.events:
                self.events[event] = []
            self.events[event].append(handler)
            return handler
        return event_ready

    def trig(self, event, *args, **kwargs):
        if event in self.events:
            for handler in self.events[event]:
                handler(*args, **kwargs)

    def send_json(self, payload):
        self.ws.send(json.dumps(payload))


    def receive_json(self):
      response = self.ws.recv()
      if response:
        return json.loads(response)
        
    def send_and_receive_json(self, payload):
      self.ws.send(json.dumps(payload))
      res = json.loads(self.ws.recv())
      return res

    def heartbeat(self):
      timer = 40
      payload = {
      "op": 1,
      "d": None
    }
      while True:
        time.sleep(timer)
        self.send_json(payload)

    def start_ws(self):
      self.ws = websocket.WebSocket()
      self.ws.connect('wss://gateway.discord.gg/?encoding=json')
      self.send_json(StartSocket(self.token))
      Thread(target=self.event_manager).start()

    def get_object(self, data):
      if data['t'] == "READY":
        self.session_id = data.get('d').get('session_id')
        if not self.client.no_warns:
          print(f"{self.client.var_} - Session ID: {self.session_id}")
        
      if data['t'] == 'MESSAGE_CREATE':
        return MessageContent(data).MessageContent

      if data['t'] == 'GUILD_MEMBER_LIST_UPDATE':
        if data['s'] == 3:
          try:
            for item in data['d']['ops'][0]['items']:
              if item.get('member'):
                self.client.online_members.append(AuthorProfile(item['member']['user']).AuthorProfile)
          except:self.client.online_members = []
      
      return data
      
    def event_manager(self):
      while True:
        event = self.receive_json()
        try:
            self.trig('all', event)
            self.trig(event['t'], self.get_object(event))

        except:
            self.trig('others', event)


import json
import time
import random
import base64
import requests
import websocket
from requests.structures import CaseInsensitiveDict
from threading import Thread
from user_discord.utils.payloads import StartSocket
from user_discord.utils.objects import MessageContent
from . import SocketDiscord
from . import discord
from .utils import youtube
import asyncio
from .utils import Thread


global_variables = {}

class ClientDiscord:

  def __init__(self, no_warns: bool = False):
    self.no_warns = no_warns
    
    self.token = None

    self.Username = None
    self.email = None
    self.id = None
    self.telephone = None

    #Guild parameters
    self.online_members = []
    
    #Reqs
    self.headers = CaseInsensitiveDict()
    self.var_ = "(user_discord)"
    


  def generate_nonce(self):
    return str(random.randint(int("1"+"0"*18), int("21"+"0"*17)))

  def login_token(self, token):
    # "Login" in account with token
    global requests
    global global_variables
    api = "https://discord.com/api/users/@me"
    temp_headers = CaseInsensitiveDict()
    temp_headers['Authorization'] = token
    res = requests.get(api, headers=temp_headers)
    if res.status_code != 200:
      return Exception(res.text)
    else:
      data = res.json()
      self.headers['Authorization'] = token
      self.web = requests.Session()
      self.web.headers = self.headers
      self.username = data['username']
      self.id = data['id']
      self.email = data['email']
      self.telephone = data['phone']
      self.token = token
      global_variables = self
      if not self.no_warns:
        print(f'{self.var_} - Logged as: {self.username}')
      return data

  def send_message(self, channelId, message):
    #Send message in channel
    api = f"https://discord.com/api/v9/channels/{channelId}/messages"
    payload = {
      "content": message,
      "flags": 0,
      "nonce": self.generate_nonce(),
      "tts": False
    }
    res = self.web.post(api, json=payload)
    if res.status_code != 200:
      return Exception(res.text)
    else:
      return res.json()
      
  def delete_message(self,channelId, messageId):
    #Delete a message by channel and message id
    api = f"https://discord.com/api/v9/channels/{channelId}/messages/{messageId}"
    res = self.web.delete(api)
    if res.status_code != 200:
      return Exception(res.text)
    else:
      return res.status_code

  def get_dms(self):
    #Get a list of users in DM
    api = "https://discordapp.com/api/users/@me/channels"
    res = self.web.get(api)
    if res.status_code != 200:
      return Exception(res.text)
    else:
      return res.json()

  def get_servers(self):
    # Get a list of your servers
    api = "https://discord.com/api/users/@me/guilds"
    res = self.web.get(api)
    if res.status_code != 200:
      return Exception(res.text)
    else:
      return res.json()

  def get_friends(self):
    # Get a list of your friends
    api = "https://discord.com/api/users/@me/relationships"
    res = self.web.get(api)
    if res.status_code != 200:
      return Exception(res.text)
    else:
      return res.json()

  def get_channel_messages(self, channelId, limit: int = 50):
    #Get a list of messages in channel
    api = f"https://discord.com/api/v9/channels/{channelId}/messages?limit={limit}"
    res = self.web.get(api)
    if res.status_code != 200:
      return Exception(res.text)
    else:
      return res.json()

  def set_status(self, status, content: str = None, emoji: str = None):
    #Set user status
    api = "https://discord.com/api/v9/users/@me/settings"
    payload = {"status": status, "custom_status":{"text":content,"emoji_name":emoji}}
    res = self.web.patch(api, json=payload)
    if res.status_code != 200:
      return Exception(res.text)
    else:
      return res.json()

  def set_bio(self, content):
    #Set user bio
    api = "https://discord.com/api/v9/users/@me/profile"
    payload = {"bio": content}
    res = self.web.patch(api, json=payload)
    if res.status_code != 200:
      return Exception(res.text)
    else:
      return res.json()

  def set_avatar(self, url):
    #Change user avatar by url
    image_data = requests.get(url).content
    image_b64 = base64.b64encode(image_data).decode('utf-8')
    data = f"data:image/png;base64,{image_b64}"

    api = "https://discord.com/api/v9/users/@me"
    payload = {"avatar": data}
    res = self.web.patch(api, json=payload)
    if res.status_code != 200:
      return Exception(res.text)
    else:
      return res.json()
      
  def get_members(self, socket, guild_id, channel_id):
    #Get members from channel
    socket.send_json({"op":14,"d":{"guild_id":guild_id,"channels":{channel_id:[[0,99]]}}})
    time.sleep(2)
    return self.online_members

  def get_user(self, user_id):
    #Get user from id
    api = "https://discord.com/api/v9/users/@me/channels"
    res = self.web.post(api, json={"recipients": [str(user_id)]})
    if res.status_code != 200:
      return Exception(res.text)
    else:
      return res.json()

  class transmit_yt_song:
    def __init__(self, channel_id, url):
      self.voice_thread = None
      self.channel_id = channel_id
      self.url = url
      
    def trans_model(self, channel_id, url):
          class VoiceClient(discord.Client):
            async def on_ready(self):
              self.vc = await player.get_channel(int(channel_id)).connect()
              source = await youtube.YTDLSource.from_url(url)
              await self.vc.play(source)
          player = VoiceClient()
          player.run(global_variables.token)
        
    def start(self):
        self.voice_thread = Thread(target=self.trans_model, args=[self.channel_id, self.url])
        self.voice_thread.start()
        
    def stop(self):
        self.voice_thread.kill()
        self.voice_thread.join()


"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

from typing import TYPE_CHECKING

from .mixins import Hashable

if TYPE_CHECKING:
    from .state import ConnectionState
    from .types.user import UserAffinity as UserAffinityPayload, GuildAffinity as GuildAffinityPayload

__all__ = (
    'UserAffinity',
    'GuildAffinity',
)


class UserAffinity(Hashable):
    """Represents a user's affinity with another user.

    User affinities that the current user has a mutual guild with are treated as implicit relationships,
    meaning that you get their user presence synced like a friend.

    These implicit relationships are not *real* relationships, and are therefore not returned by the API.
    However, they are lazily added to cache by the library when detected.

    .. container:: operations

        .. describe:: x == y

            Checks if two affinities are equal.

        .. describe:: x != y

            Checks if two affinities are not equal.

        .. describe:: hash(x)

            Return the affinity's hash.

    .. versionadded:: 2.0

    Attributes
    ----------
    user_id: :class:`int`
        The ID of the user being compared.
    affinity: :class:`float`
        The affinity score.
    """

    __slots__ = ('_state', 'user_id', 'affinity')

    def __init__(self, *, state: ConnectionState, data: UserAffinityPayload):
        self._state = state
        self.user_id = int(data['user_id'])
        self.affinity = data['affinity']

    def __repr__(self) -> str:
        return f'<UserAffinity user_id={self.user_id} affinity={self.affinity}>'

    @property
    def id(self) -> int:
        """:class:`int`: The ID of the user being compared."""
        return self.user_id

    @property
    def user(self):
        """Optional[:class:`User`]: The user being compared."""
        return self._state.get_user(self.user_id)


class GuildAffinity(Hashable):
    """Represents a user's affinity with a guild.

    .. container:: operations

        .. describe:: x == y

            Checks if two affinities are equal.

        .. describe:: x != y

            Checks if two affinities are not equal.

        .. describe:: hash(x)

            Return the affinity's hash.

    .. versionadded:: 2.0

    Attributes
    ----------
    guild_id: :class:`int`
        The ID of the guild being compared.
    affinity: :class:`float`
        The affinity score.
    """

    __slots__ = ('_state', 'guild_id', 'affinity')

    def __init__(self, *, state: ConnectionState, data: GuildAffinityPayload):
        self._state = state
        self.guild_id = int(data['guild_id'])
        self.affinity = data['affinity']

    def __repr__(self) -> str:
        return f'<GuildAffinity guild_id={self.guild_id} affinity={self.affinity}>'

    @property
    def id(self) -> int:
        """:class:`int`: The ID of the guild being compared."""
        return self.guild_id

    @property
    def guild(self):
        """Optional[:class:`Guild`]: The guild being compared."""
        return self._state._get_guild(self.guild_id)



def StartSocket(token):
  return {
      "op": 2,
      "d": {
          "token": token,
          "capabilities": 8189,
          "properties": {
              "os": "Windows",
              "browser": "Chrome",
              "device": "",
              "system_locale": "pt-BR",
              "browser_user_agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/114.0.0.0 Safari/537.36",
              "browser_version": "114.0.0.0",
              "os_version": "10",
              "referrer": "",
              "referring_domain": "",
              "referrer_current": "",
              "referring_domain_current": "",
              "release_channel": "stable",
              "client_build_number": 208319,
              "client_event_source": None
          }
      }
  }



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Dict, List, Literal, Optional, TypedDict, Union
from typing_extensions import NotRequired

from .guild import PartialGuild
from .snowflake import Snowflake
from .team import Team
from .user import PartialUser


class BaseApplication(TypedDict):
    id: Snowflake
    name: str
    description: str
    icon: Optional[str]
    cover_image: NotRequired[Optional[str]]
    type: Optional[int]
    primary_sku_id: NotRequired[Snowflake]
    summary: NotRequired[Literal['']]


class MetadataApplication(BaseApplication):
    bot: NotRequired[PartialUser]


class IntegrationApplication(MetadataApplication):
    role_connections_verification_url: NotRequired[Optional[str]]


class PartialApplication(BaseApplication):
    owner: NotRequired[PartialUser]  # Not actually ever present in partial app
    team: NotRequired[Team]
    verify_key: str
    description: str
    cover_image: NotRequired[Optional[str]]
    flags: NotRequired[int]
    rpc_origins: NotRequired[List[str]]
    hook: NotRequired[bool]
    overlay: NotRequired[bool]
    overlay_compatibility_hook: NotRequired[bool]
    terms_of_service_url: NotRequired[str]
    privacy_policy_url: NotRequired[str]
    max_participants: NotRequired[Optional[int]]
    bot_public: NotRequired[bool]
    bot_require_code_grant: NotRequired[bool]
    integration_public: NotRequired[bool]
    integration_require_code_grant: NotRequired[bool]
    guild_id: NotRequired[Snowflake]
    primary_sku_id: NotRequired[Snowflake]
    slug: NotRequired[str]
    developers: NotRequired[List[Company]]
    publishers: NotRequired[List[Company]]
    aliases: NotRequired[List[str]]
    eula_id: NotRequired[Snowflake]
    embedded_activity_config: NotRequired[EmbeddedActivityConfig]
    guild: NotRequired[PartialGuild]
    install_params: NotRequired[ApplicationInstallParams]


class ApplicationDiscoverability(TypedDict):
    discoverability_state: int
    discovery_eligibility_flags: int


class Application(PartialApplication, IntegrationApplication, ApplicationDiscoverability):
    redirect_uris: List[str]
    interactions_endpoint_url: Optional[str]
    verification_state: int
    store_application_state: int
    rpc_application_state: int
    creator_monetization_state: int
    role_connections_verification_url: NotRequired[Optional[str]]


class WhitelistedUser(TypedDict):
    user: PartialUser
    state: Literal[1, 2]


class Asset(TypedDict):
    id: Snowflake
    name: str
    type: int


class StoreAsset(TypedDict):
    id: Snowflake
    size: int
    width: int
    height: int
    mime_type: str


class Company(TypedDict):
    id: Snowflake
    name: str


class EULA(TypedDict):
    id: Snowflake
    name: str
    content: str


class BaseAchievement(TypedDict):
    id: Snowflake
    name: Union[str, Dict[str, Union[str, Dict[str, str]]]]
    name_localizations: NotRequired[Dict[str, str]]
    description: Union[str, Dict[str, Union[str, Dict[str, str]]]]
    description_localizations: NotRequired[Dict[str, str]]
    icon_hash: str
    secure: bool
    secret: bool


class Achievement(BaseAchievement):
    application_id: Snowflake


class Ticket(TypedDict):
    ticket: str


class Branch(TypedDict):
    id: Snowflake
    live_build_id: NotRequired[Optional[Snowflake]]
    created_at: NotRequired[str]
    name: NotRequired[str]


class BranchSize(TypedDict):
    size_kb: str  # Stringified float


class DownloadSignature(TypedDict):
    endpoint: str
    expires: int
    signature: str


class Build(TypedDict):
    application_id: NotRequired[Snowflake]
    created_at: NotRequired[str]
    id: Snowflake
    manifests: List[Manifest]
    status: Literal['CORRUPTED', 'INVALID', 'READY', 'VALIDATING', 'UPLOADED', 'UPLOADING', 'CREATED']
    source_build_id: NotRequired[Optional[Snowflake]]
    version: NotRequired[Optional[str]]


class CreatedBuild(TypedDict):
    build: Build
    manifest_uploads: List[Manifest]


class BuildFile(TypedDict):
    id: Snowflake
    md5_hash: NotRequired[str]


class CreatedBuildFile(TypedDict):
    id: str
    url: str


class ManifestLabel(TypedDict):
    application_id: Snowflake
    id: Snowflake
    name: NotRequired[str]


class Manifest(TypedDict):
    id: Snowflake
    label: ManifestLabel
    redistributable_label_ids: NotRequired[List[Snowflake]]
    url: Optional[str]


class ActivityStatistics(TypedDict):
    application_id: NotRequired[Snowflake]
    user_id: NotRequired[Snowflake]
    total_duration: int
    total_discord_sku_duration: NotRequired[int]
    last_played_at: str


class GlobalActivityStatistics(TypedDict):
    application_id: Snowflake
    user_id: Snowflake
    duration: int
    updated_at: str


EmbeddedActivityPlatform = Literal['web', 'android', 'ios']


class ClientPlatformConfig(TypedDict):
    label_type: int
    label_until: Optional[str]
    release_phase: str


class EmbeddedActivityConfig(TypedDict):
    activity_preview_video_asset_id: NotRequired[Optional[Snowflake]]
    client_platform_config: Dict[EmbeddedActivityPlatform, ClientPlatformConfig]
    default_orientation_lock_state: Literal[1, 2, 3]
    tablet_default_orientation_lock_state: Literal[1, 2, 3]
    free_period_ends_at: NotRequired[Optional[str]]
    free_period_starts_at: NotRequired[Optional[str]]
    premium_tier_requirement: NotRequired[Optional[Literal[1, 2, 3]]]
    requires_age_gate: bool
    shelf_rank: int
    supported_platforms: List[EmbeddedActivityPlatform]


class ApplicationInstallParams(TypedDict):
    scopes: List[str]
    permissions: int


class ActiveDeveloperWebhook(TypedDict):
    channel_id: Snowflake
    webhook_id: Snowflake


class ActiveDeveloperResponse(TypedDict):
    follower: ActiveDeveloperWebhook


class RoleConnectionMetadata(TypedDict):
    type: Literal[1, 2, 3, 4, 5, 6, 7, 8]
    key: str
    name: str
    description: str
    name_localizations: NotRequired[Dict[str, str]]
    description_localizations: NotRequired[Dict[str, str]]


class PartialRoleConnection(TypedDict):
    platform_name: Optional[str]
    platform_username: Optional[str]
    metadata: Dict[str, str]


class RoleConnection(PartialRoleConnection):
    application: MetadataApplication
    application_metadata: List[RoleConnectionMetadata]



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import Any, Dict, Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from aiohttp import BasicAuth

# fmt: off
__all__ = (
    'CaptchaHandler',
)
# fmt: on


class CaptchaHandler:
    """A class that represents a captcha handler.

    This class allows you to implement a protocol to solve captchas required by Discord.
    This is an abstract class. The library provides no concrete implementation.

    These classes are passed to :class:`Client`.
    """

    async def startup(self):
        """|coro|

        An abstract method that is called by the library at startup.

        This is meant to provide an async startup method for the handler.
        This isn't guaranteed to be run once.

        The default implementation does nothing.
        """
        pass

    async def prefetch_token(self, proxy: Optional[str], proxy_auth: Optional[BasicAuth], /) -> None:
        """|coro|

        An abstract method that is called a bit before a captcha token is required.
        Not guaranteed to be called.

        It's meant to signal the handler to begin preparing for the fetching of a token, if applicable.
        Keep in mind that Discord has multiple captcha sitekeys.

        The default implementation does nothing.

        Parameters
        ----------
        proxy: Optional[:class:`str`]
            The current proxy of the client.
        proxy_auth: Optional[:class:`aiohttp.BasicAuth`]
            The proxy's auth.
        """
        pass

    async def fetch_token(
        self,
        data: Dict[str, Any],
        proxy: Optional[str],
        proxy_auth: Optional[BasicAuth],
        /,
    ) -> str:
        """|coro|

        An abstract method that is called to fetch a captcha token.

        If there is no token available, it should wait until one is
        generated before returning.

        Parameters
        ------------
        data: Dict[:class:`str`, :class:`Any`]
            The raw error from Discord containing the captcha info.
        proxy: Optional[:class:`str`]
            The current proxy of the client.
        proxy_auth: Optional[:class:`aiohttp.BasicAuth`]
            The proxy's auth.

        Returns
        --------
        :class:`str`
            A captcha token.
        """
        raise NotImplementedError



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from .billing import PaymentSource
from .enums import (
    PaymentGateway,
    SubscriptionDiscountType,
    SubscriptionInterval,
    SubscriptionInvoiceStatus,
    SubscriptionStatus,
    SubscriptionType,
    try_enum,
)
from .metadata import Metadata
from .mixins import Hashable
from .utils import MISSING, _get_as_snowflake, parse_time, snowflake_time, utcnow

if TYPE_CHECKING:
    from typing_extensions import Self

    from .abc import Snowflake
    from .guild import Guild
    from .state import ConnectionState
    from .types.subscriptions import (
        SubscriptionDiscount as SubscriptionDiscountPayload,
        SubscriptionInvoice as SubscriptionInvoicePayload,
        SubscriptionInvoiceItem as SubscriptionInvoiceItemPayload,
        SubscriptionItem as SubscriptionItemPayload,
        SubscriptionRenewalMutations as SubscriptionRenewalMutationsPayload,
        PartialSubscription as PartialSubscriptionPayload,
        Subscription as SubscriptionPayload,
        SubscriptionTrial as SubscriptionTrialPayload,
    )

__all__ = (
    'SubscriptionItem',
    'SubscriptionDiscount',
    'SubscriptionInvoiceItem',
    'SubscriptionInvoice',
    'SubscriptionRenewalMutations',
    'Subscription',
    'SubscriptionTrial',
)


class SubscriptionItem(Hashable):
    """Represents a Discord subscription item.

    .. container:: operations

        .. describe:: x == y

            Checks if two subscription items are equal.

        .. describe:: x != y

            Checks if two subscription items are not equal.

        .. describe:: hash(x)

            Returns the item's hash.

        .. describe:: len(x)

            Returns the quantity of the subscription item.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: Optional[:class:`int`]
        The ID of the subscription item. Always available when received from the API.
    quantity: :class:`int`
        How many of the item have been/are being purchased.
    plan_id: :class:`int`
        The ID of the plan the item is for.
    """

    __slots__ = ('id', 'quantity', 'plan_id')

    def __init__(self, *, id: Optional[int] = None, plan_id: int, quantity: int = 1) -> None:
        self.id: Optional[int] = id
        self.quantity: int = quantity
        self.plan_id: int = plan_id

    def __repr__(self) -> str:
        return f'<SubscriptionItem {f"id={self.id} " if self.id else ""}plan_id={self.plan_id} quantity={self.quantity}>'

    def __len__(self) -> int:
        return self.quantity

    @classmethod
    def from_dict(cls, data: SubscriptionItemPayload) -> Self:
        return cls(id=int(data['id']), plan_id=int(data['plan_id']), quantity=int(data.get('quantity', 1)))

    def to_dict(self, with_id: bool = True) -> dict:
        data = {
            'quantity': self.quantity,
            'plan_id': self.plan_id,
        }
        if self.id and with_id:
            data['id'] = self.id

        return data


class SubscriptionDiscount:
    """Represents a discount on a Discord subscription item.

    .. container:: operations

        .. describe:: int(x)

            Returns the discount's amount.

    .. versionadded:: 2.0

    Attributes
    ----------
    type: :class:`SubscriptionDiscountType`
        The type of the discount.
    amount: :class:`int`
        How much the discount is.
    """

    __slots__ = ('type', 'amount')

    def __init__(self, data: SubscriptionDiscountPayload) -> None:
        self.type: SubscriptionDiscountType = try_enum(SubscriptionDiscountType, data['type'])
        self.amount: int = data['amount']

    def __repr__(self) -> str:
        return f'<SubscriptionDiscount type={self.type!r} amount={self.amount}>'

    def __int__(self) -> int:
        return self.amount


class SubscriptionInvoiceItem(Hashable):
    """Represents an invoice item.

    .. container:: operations

        .. describe:: x == y

            Checks if two invoice items are equal.

        .. describe:: x != y

            Checks if two invoice items are not equal.

        .. describe:: hash(x)

            Returns the invoice's hash.

        .. describe:: len(x)

            Returns the quantity of the invoice item.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The ID of the invoice item.
    quantity: :class:`int`
        How many of the item have been/are being purchased.
    amount: :class:`int`
        The price of the item. This includes discounts.
    proration: :class:`bool`
        Whether the item is prorated.
    plan_id: :class:`int`
        The ID of the subscription plan the item represents.
    plan_price: :class:`int`
        The price of the subscription plan the item represents. This does not include discounts.
    discounts: List[:class:`SubscriptionDiscount`]
        A list of discounts applied to the item.
    """

    __slots__ = ('id', 'quantity', 'amount', 'proration', 'plan_id', 'plan_price', 'discounts')

    def __init__(self, data: SubscriptionInvoiceItemPayload) -> None:
        self.id: int = int(data['id'])
        self.quantity: int = data['quantity']
        self.amount: int = data['amount']
        self.proration: bool = data.get('proration', False)
        self.plan_id: int = int(data['subscription_plan_id'])
        self.plan_price: int = data['subscription_plan_price']
        self.discounts: List[SubscriptionDiscount] = [SubscriptionDiscount(d) for d in data['discounts']]

    def __repr__(self) -> str:
        return f'<SubscriptionInvoiceItem id={self.id} quantity={self.quantity} amount={self.amount}>'

    def __len__(self) -> int:
        return self.quantity

    @property
    def savings(self) -> int:
        """:class:`int`: The total amount of discounts on the invoice item."""
        return self.plan_price - self.amount

    def is_discounted(self) -> bool:
        """:class:`bool`: Indicates if the invoice item has a discount."""
        return bool(self.discounts)

    def is_trial(self) -> bool:
        """:class:`bool`: Indicates if the invoice item is a trial."""
        return not self.amount or any(discount.type is SubscriptionDiscountType.premium_trial for discount in self.discounts)


class SubscriptionInvoice(Hashable):
    """Represents an invoice for a Discord subscription.

    .. container:: operations

        .. describe:: x == y

            Checks if two invoices are equal.

        .. describe:: x != y

            Checks if two invoices are not equal.

        .. describe:: hash(x)

            Returns the invoice's hash.

    .. versionadded:: 2.0

    Attributes
    ----------
    subscription: Optional[:class:`Subscription`]
        The subscription the invoice is for. Not available for new subscription previews.
    id: :class:`int`
        The ID of the invoice.
    status: Optional[:class:`SubscriptionInvoiceStatus`]
        The status of the invoice. Not available for subscription previews.
    currency: :class:`str`
        The currency the invoice is in.
    subtotal: :class:`int`
        The subtotal of the invoice.
    tax: :class:`int`
        The tax applied to the invoice.
    total: :class:`int`
        The total of the invoice.
    tax_inclusive: :class:`bool`
        Whether the subtotal is inclusive of all taxes.
    items: List[:class:`SubscriptionInvoiceItem`]
        The items in the invoice.
    current_period_start: :class:`datetime.datetime`
        When the current billing period started.
    current_period_end: :class:`datetime.datetime`
        When the current billing period ends.
    """

    __slots__ = (
        '_state',
        'subscription',
        'id',
        'status',
        'currency',
        'subtotal',
        'tax',
        'total',
        'tax_inclusive',
        'items',
        'current_period_start',
        'current_period_end',
    )

    def __init__(
        self, subscription: Optional[Subscription], *, data: SubscriptionInvoicePayload, state: ConnectionState
    ) -> None:
        self._state = state
        self.subscription = subscription
        self._update(data)

    def _update(self, data: SubscriptionInvoicePayload) -> None:
        self.id: int = int(data['id'])
        self.status: Optional[SubscriptionInvoiceStatus] = (
            try_enum(SubscriptionInvoiceStatus, data['status']) if 'status' in data else None
        )
        self.currency: str = data['currency']
        self.subtotal: int = data['subtotal']
        self.tax: int = data.get('tax', 0)
        self.total: int = data['total']
        self.tax_inclusive: bool = data['tax_inclusive']
        self.items: List[SubscriptionInvoiceItem] = [SubscriptionInvoiceItem(d) for d in data.get('invoice_items', [])]

        self.current_period_start: datetime = parse_time(data['subscription_period_start'])  # type: ignore # Should always be a datetime
        self.current_period_end: datetime = parse_time(data['subscription_period_end'])  # type: ignore # Should always be a datetime

    def __repr__(self) -> str:
        return f'<SubscriptionInvoice id={self.id} status={self.status!r} total={self.total}>'

    def is_discounted(self) -> bool:
        """:class:`bool`: Indicates if the invoice has a discount."""
        return any(item.discounts for item in self.items)

    def is_preview(self) -> bool:
        """:class:`bool`: Indicates if the invoice is a preview and not real."""
        return self.subscription is None or self.status is None

    async def pay(
        self,
        payment_source: Optional[Snowflake] = None,
        currency: str = 'usd',
        *,
        payment_source_token: Optional[str] = None,
        return_url: Optional[str] = None,
    ) -> None:
        """|coro|

        Pays the invoice.

        Parameters
        ----------
        payment_source: Optional[:class:`PaymentSource`]
            The payment source the invoice should be paid with.
        currency: :class:`str`
            The currency to pay with.
        payment_source_token: Optional[:class:`str`]
            The token used to authorize with the payment source.
        return_url: Optional[:class:`str`]
            The URL to return to after the payment is complete.

        Raises
        ------
        TypeError
            The invoice is a preview and not real.
        NotFound
            The invoice is not open or found.
        HTTPException
            Paying the invoice failed.
        """
        if self.is_preview() or not self.subscription:
            raise TypeError('Cannot pay a nonexistant invoice')

        data = await self._state.http.pay_invoice(
            self.subscription.id,
            self.id,
            payment_source.id if payment_source else None,
            payment_source_token,
            currency,
            return_url,
        )
        self.subscription._update(data)


class SubscriptionRenewalMutations:
    """Represents a subscription renewal mutation.

    This represents changes to a subscription that will occur after renewal.

    .. container:: operations

        .. describe:: len(x)

            Returns the number of items in the changed subscription, including quantity.

        .. describe:: bool(x)

            Returns whether any mutations are present.

    .. versionadded:: 2.0

    Attributes
    ----------
    payment_gateway_plan_id: Optional[:class:`str`]
        The payment gateway's new plan ID for the subscription.
        This signifies an external plan change.
    items: Optional[List[:class:`SubscriptionItem`]]
        The new items of the subscription.
    """

    __slots__ = ('payment_gateway_plan_id', 'items')

    def __init__(self, data: SubscriptionRenewalMutationsPayload) -> None:
        self.payment_gateway_plan_id: Optional[str] = data.get('payment_gateway_plan_id')
        self.items: Optional[List[SubscriptionItem]] = (
            [SubscriptionItem.from_dict(item) for item in data['items']] if 'items' in data else None
        )

    def __repr__(self) -> str:
        return (
            f'<SubscriptionRenewalMutations payment_gateway_plan_id={self.payment_gateway_plan_id!r} items={self.items!r}>'
        )

    def __len__(self) -> int:
        return sum(item.quantity for item in self.items) if self.items else 0

    def __bool__(self) -> bool:
        return self.is_mutated()

    def is_mutated(self) -> bool:
        """:class:`bool`: Checks if any renewal mutations exist."""
        return self.payment_gateway_plan_id is not None or self.items is not None


class Subscription(Hashable):
    """Represents a Discord subscription.

    .. container:: operations

        .. describe:: x == y

            Checks if two premium subscriptions are equal.

        .. describe:: x != y

            Checks if two premium subscriptions are not equal.

        .. describe:: hash(x)

            Returns the subscription's hash.

        .. describe:: len(x)

            Returns the number of items in the subscription, including quantity.

        .. describe:: bool(x)

            Checks if the subscription is currently active and offering perks.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The ID of the subscription.
    type: :class:`SubscriptionType`
        The type of the subscription.
    status: Optional[:class:`SubscriptionStatus`]
        The status of the subscription. This is ``None`` for fake subscriptions.
    payment_gateway: Optional[:class:`PaymentGateway`]
        The payment gateway used to bill the subscription.
    currency: :class:`str`
        The currency the subscription is billed in.
    items: List[:class:`SubscriptionItem`]
        The items in the subscription.
    renewal_mutations: :class:`SubscriptionRenewalMutations`
        The mutations to the subscription that will occur after renewal.
    trial_id: Optional[:class:`int`]
        The ID of the trial the subscription is from, if applicable.
    payment_source_id: Optional[:class:`int`]
        The ID of the payment source the subscription is paid with, if applicable.
    payment_gateway_plan_id: Optional[:class:`str`]
        The payment gateway's plan ID for the subscription, if applicable.
    payment_gateway_subscription_id: Optional[:class:`str`]
        The payment gateway's subscription ID for the subscription, if applicable.
    created_at: :class:`datetime.datetime`
        When the subscription was created.
    canceled_at: Optional[:class:`datetime.datetime`]
        When the subscription was canceled.
        This is only available for subscriptions with a :attr:`status` of :attr:`SubscriptionStatus.canceled`.
    current_period_start: :class:`datetime.datetime`
        When the current billing period started.
    current_period_end: :class:`datetime.datetime`
        When the current billing period ends.
    trial_ends_at: Optional[:class:`datetime.datetime`]
        When the trial ends, if applicable.
    streak_started_at: Optional[:class:`datetime.datetime`]
        When the current subscription streak started.
    ended_at: Optional[:class:`datetime.datetime`]
        When the subscription finally ended.
    metadata: :class:`Metadata`
        Extra metadata about the subscription.
    latest_invoice: Optional[:class:`SubscriptionInvoice`]
        The latest invoice for the subscription, if applicable.
    """

    __slots__ = (
        '_state',
        'id',
        'type',
        'status',
        'payment_gateway',
        'currency',
        'items',
        'renewal_mutations',
        'trial_id',
        'payment_source_id',
        'payment_gateway_plan_id',
        'payment_gateway_subscription_id',
        'created_at',
        'canceled_at',
        'current_period_start',
        'current_period_end',
        'trial_ends_at',
        'streak_started_at',
        'ended_at',
        'metadata',
        'latest_invoice',
    )

    def __init__(self, *, data: Union[PartialSubscriptionPayload, SubscriptionPayload], state: ConnectionState) -> None:
        self._state = state
        self._update(data)

    def __repr__(self) -> str:
        return f'<Subscription id={self.id} currency={self.currency!r} items={self.items!r}>'

    def __len__(self) -> int:
        return sum(item.quantity for item in self.items)

    def __bool__(self) -> bool:
        return self.is_active()

    def _update(self, data: PartialSubscriptionPayload) -> None:
        self.id: int = int(data['id'])
        self.type: SubscriptionType = try_enum(SubscriptionType, data['type'])
        self.status: Optional[SubscriptionStatus] = (
            try_enum(SubscriptionStatus, data['status']) if 'status' in data else None  # type: ignore # ???
        )
        self.payment_gateway: Optional[PaymentGateway] = (
            try_enum(PaymentGateway, data['payment_gateway']) if 'payment_gateway' in data else None
        )
        self.currency: str = data.get('currency', 'usd')
        self.items: List[SubscriptionItem] = [SubscriptionItem.from_dict(item) for item in data.get('items', [])]
        self.renewal_mutations: SubscriptionRenewalMutations = SubscriptionRenewalMutations(
            data.get('renewal_mutations') or {}
        )

        self.trial_id: Optional[int] = _get_as_snowflake(data, 'trial_id')
        self.payment_source_id: Optional[int] = _get_as_snowflake(data, 'payment_source_id')
        self.payment_gateway_plan_id: Optional[str] = data.get('payment_gateway_plan_id')
        self.payment_gateway_subscription_id: Optional[str] = data.get('payment_gateway_subscription_id')

        self.created_at: datetime = parse_time(data.get('created_at')) or snowflake_time(self.id)
        self.canceled_at: Optional[datetime] = parse_time(data.get('canceled_at'))

        self.current_period_start: datetime = parse_time(data['current_period_start'])
        self.current_period_end: datetime = parse_time(data['current_period_end'])
        self.trial_ends_at: Optional[datetime] = parse_time(data.get('trial_ends_at'))
        self.streak_started_at: Optional[datetime] = parse_time(data.get('streak_started_at'))

        metadata = data.get('metadata') or {}
        self.ended_at: Optional[datetime] = parse_time(metadata.get('ended_at', None))
        self.metadata: Metadata = Metadata(metadata)

        self.latest_invoice: Optional[SubscriptionInvoice] = (
            SubscriptionInvoice(self, data=data['latest_invoice'], state=self._state) if 'latest_invoice' in data else None  # type: ignore # ???
        )

    @property
    def cancelled_at(self) -> Optional[datetime]:
        """Optional[:class:`datetime.datetime`]: When the subscription was canceled.
        This is only available for subscriptions with a :attr:`status` of :attr:`SubscriptionStatus.canceled`.

        This is an alias of :attr:`canceled_at`.
        """
        return self.canceled_at

    @property
    def guild(self) -> Optional[Guild]:
        """:class:`Guild`: The guild the subscription's entitlements apply to, if applicable."""
        return self._state._get_guild(self.metadata.guild_id)

    @property
    def grace_period(self) -> int:
        """:class:`int`: How many days past the renewal date the user has available to pay outstanding invoices.

        .. note::

            This is a static value and does not change based on the subscription's status.
            For that, see :attr:`remaining`.
        """
        return 7 if self.payment_source_id else 3

    @property
    def remaining(self) -> timedelta:
        """:class:`datetime.timedelta`: The remaining time until the subscription ends."""
        if self.status in (SubscriptionStatus.active, SubscriptionStatus.cancelled):
            return self.current_period_end - utcnow()
        elif self.status == SubscriptionStatus.past_due:
            if self.payment_gateway == PaymentGateway.google and self.metadata.google_grace_period_expires_date:
                return self.metadata.google_grace_period_expires_date - utcnow()
            return (self.current_period_start + timedelta(days=self.grace_period)) - utcnow()
        elif self.status == SubscriptionStatus.account_hold:
            # Max hold time is 30 days
            return (self.current_period_start + timedelta(days=30)) - utcnow()
        return timedelta()

    @property
    def trial_remaining(self) -> timedelta:
        """:class:`datetime.timedelta`: The remaining time until the trial applied to the subscription ends."""
        if not self.trial_id:
            return timedelta()
        if not self.trial_ends_at:
            # Infinite trial?
            return self.remaining
        return self.trial_ends_at - utcnow()

    def is_active(self) -> bool:
        """:class:`bool`: Indicates if the subscription is currently active and providing perks."""
        return self.remaining > timedelta()

    def is_trial(self) -> bool:
        """:class:`bool`: Indicates if the subscription is a trial."""
        return self.trial_id is not None

    async def edit(
        self,
        items: List[SubscriptionItem] = MISSING,
        payment_source: Snowflake = MISSING,
        currency: str = MISSING,
        *,
        status: SubscriptionStatus = MISSING,
        payment_source_token: Optional[str] = None,
    ) -> None:
        """|coro|

        Edits the subscription.

        All parameters are optional.

        Parameters
        ----------
        items: List[:class:`SubscriptionItem`]
            The new subscription items to use.
        payment_source: :class:`int`
            The new payment source for payment.
        currency: :class:`str`
            The new currency to use for payment.
        status: :class:`SubscriptionStatus`
            The new status of the subscription.
        payment_source_token: Optional[:class:`str`]
            The token used to authorize with the payment source.

        Raises
        ------
        Forbidden
            You do not have permissions to edit the subscription.
        HTTPException
            Editing the subscription failed.
        """
        payload = {}
        if items is not MISSING:
            payload['items'] = [item.to_dict() for item in items] if items else []
        if payment_source is not MISSING:
            payload['payment_source_id'] = payment_source.id
            payload['payment_source_token'] = payment_source_token
        if currency is not MISSING:
            payload['currency'] = currency
        if status is not MISSING:
            payload['status'] = int(status)

        data = await self._state.http.edit_subscription(self.id, **payload)
        self._update(data)

    async def delete(self) -> None:
        """|coro|

        Deletes the subscription.

        There is an alias of this called :meth:`cancel`.

        Raises
        ------
        HTTPException
            Deleting the subscription failed.
        """
        await self._state.http.delete_subscription(self.id)

    async def cancel(self) -> None:
        """|coro|

        Deletes the subscription.

        Alias of :meth:`delete`.

        Raises
        ------
        HTTPException
            Deleting the subscription failed.
        """
        await self.delete()

    async def preview_invoice(
        self,
        *,
        items: List[SubscriptionItem] = MISSING,
        payment_source: Snowflake = MISSING,
        currency: str = MISSING,
        apply_entitlements: bool = MISSING,
        renewal: bool = MISSING,
    ) -> SubscriptionInvoice:
        """|coro|

        Preview an invoice for the subscription with the given parameters.

        All parameters are optional and default to the current subscription values.

        Parameters
        ----------
        items: List[:class:`SubscriptionItem`]
            The items the previewed invoice should have.
        payment_source: :class:`.PaymentSource`
            The payment source the previewed invoice should be paid with.
        currency: :class:`str`
            The currency the previewed invoice should be paid in.
        apply_entitlements: :class:`bool`
            Whether to apply entitlements (credits) to the previewed invoice.
        renewal: :class:`bool`
            Whether the previewed invoice should be a renewal.

        Raises
        ------
        HTTPException
            Failed to preview the invoice.

        Returns
        -------
        :class:`SubscriptionInvoice`
            The previewed invoice.
        """
        payload: Dict[str, Any] = {}
        if items is not MISSING:
            payload['items'] = [item.to_dict() for item in items] if items else []
        if payment_source:
            payload['payment_source_id'] = payment_source.id
        if currency:
            payload['currency'] = currency
        if apply_entitlements is not MISSING:
            payload['apply_entitlements'] = apply_entitlements
        if renewal is not MISSING:
            payload['renewal'] = renewal

        if payload:
            data = await self._state.http.preview_subscription_update(self.id, **payload)
        else:
            data = await self._state.http.get_subscription_preview(self.id)

        return SubscriptionInvoice(self, data=data, state=self._state)

    async def payment_source(self) -> Optional[PaymentSource]:
        """|coro|

        Retrieves the payment source the subscription is paid with, if applicable.

        Raises
        ------
        NotFound
            The payment source could not be found.
        HTTPException
            Retrieving the payment source failed.

        Returns
        -------
        Optional[:class:`PaymentSource`]
            The payment source the subscription is paid with, if applicable.
        """
        if not self.payment_source_id:
            return

        data = await self._state.http.get_payment_source(self.payment_source_id)
        return PaymentSource(data=data, state=self._state)

    async def invoices(self):
        """|coro|

        Retrieves all invoices for the subscription.

        Raises
        ------
        NotFound
            The payment source or invoices could not be found.
        HTTPException
            Retrieving the invoices failed.

        Returns
        -------
        List[:class:`SubscriptionInvoice`]
            The invoices.
        """
        state = self._state
        data = await state.http.get_subscription_invoices(self.id)
        return [SubscriptionInvoice(self, data=d, state=state) for d in data]


class SubscriptionTrial(Hashable):
    """Represents a subscription trial.

    .. container:: operations

        .. describe:: x == y

            Checks if two trials are equal.

        .. describe:: x != y

            Checks if two trials are not equal.

        .. describe:: hash(x)

            Returns the trial's hash.

    .. versionadded:: 2.0

    Attributes
    ----------
    id: :class:`int`
        The ID of the trial.
    interval: :class:`SubscriptionInterval`
        The interval of the trial.
    interval_count: :class:`int`
        How many counts of the interval the trial provides.
    """

    __slots__ = ('id', 'interval', 'interval_count', 'sku_id')

    _INTERVAL_TABLE = {
        SubscriptionInterval.day: 1,
        SubscriptionInterval.month: 30,
        SubscriptionInterval.year: 365,
    }

    def __init__(self, data: SubscriptionTrialPayload):
        self.id: int = int(data['id'])
        self.interval: SubscriptionInterval = try_enum(SubscriptionInterval, data['interval'])
        self.interval_count: int = data['interval_count']
        self.sku_id: int = int(data['sku_id'])

    def __repr__(self) -> str:
        return (
            f'<SubscriptionTrial id={self.id} interval={self.interval} '
            f'interval_count={self.interval_count} sku_id={self.sku_id}>'
        )

    @property
    def duration(self) -> timedelta:
        """:class:`datetime.timedelta`: How long the trial lasts."""
        return timedelta(days=self.interval_count * self._INTERVAL_TABLE[self.interval])



"""
The MIT License (MIT)

Copyright (c) 2021-present Dolfies

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""
from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from .enums import ConnectionType, try_enum
from .integrations import Integration
from .metadata import Metadata
from .utils import MISSING

if TYPE_CHECKING:
    from .guild import Guild
    from .state import ConnectionState
    from .types.integration import ConnectionIntegration as IntegrationPayload
    from .types.user import Connection as ConnectionPayload, PartialConnection as PartialConnectionPayload

__all__ = (
    'PartialConnection',
    'Connection',
)


class PartialConnection:
    """Represents a partial Discord profile connection.

    This is the info you get for other users' connections.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: x == y

            Checks if two connections are equal.

        .. describe:: x != y

            Checks if two connections are not equal.

        .. describe:: hash(x)

            Return the connection's hash.

        .. describe:: str(x)

            Returns the connection's name.

    Attributes
    ----------
    id: :class:`str`
        The connection's account ID.
    name: :class:`str`
        The connection's account name.
    type: :class:`ConnectionType`
        The connection service type (e.g. youtube, twitch, etc.).
    verified: :class:`bool`
        Whether the connection is verified.
    visible: :class:`bool`
        Whether the connection is visible on the user's profile.
    metadata: Optional[:class:`Metadata`]
        Various metadata about the connection.

        The contents of this are always subject to change.
    """

    __slots__ = ('id', 'name', 'type', 'verified', 'visible', 'metadata')

    def __init__(self, data: PartialConnectionPayload):
        self._update(data)

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__} id={self.id!r} name={self.name!r} type={self.type!r} visible={self.visible}>'

    def __hash__(self) -> int:
        return hash((self.type.value, self.id))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, PartialConnection):
            return self.id == other.id and self.name == other.name
        return False

    def __ne__(self, other: object) -> bool:
        if isinstance(other, PartialConnection):
            return self.id != other.id or self.name != other.name
        return True

    def _update(self, data: PartialConnectionPayload):
        self.id: str = data['id']
        self.name: str = data['name']
        self.type: ConnectionType = try_enum(ConnectionType, data['type'])
        self.verified: bool = data['verified']
        self.visible: bool = True  # If we have a partial connection, it's visible

        self.metadata: Optional[Metadata] = Metadata(data['metadata']) if 'metadata' in data else None

    @property
    def url(self) -> Optional[str]:
        """Optional[:class:`str`]: Returns a URL linking to the connection's profile, if available."""
        if self.type == ConnectionType.twitch:
            return f'https://www.twitch.tv/{self.name}'
        elif self.type == ConnectionType.youtube:
            return f'https://www.youtube.com/{self.id}'
        elif self.type == ConnectionType.skype:
            return f'skype:{self.id}?userinfo'
        elif self.type == ConnectionType.steam:
            return f'https://steamcommunity.com/profiles/{self.id}'
        elif self.type == ConnectionType.reddit:
            return f'https://www.reddit.com/u/{self.name}'
        elif self.type == ConnectionType.facebook:
            return f'https://www.facebook.com/{self.name}'
        elif self.type == ConnectionType.twitter:
            return f'https://twitter.com/{self.name}'
        elif self.type == ConnectionType.spotify:
            return f'https://open.spotify.com/user/{self.id}'
        elif self.type == ConnectionType.xbox:
            return f'https://account.xbox.com/en-US/Profile?Gamertag={self.name}'
        elif self.type == ConnectionType.github:
            return f'https://github.com/{self.name}'
        elif self.type == ConnectionType.tiktok:
            return f'https://tiktok.com/@{self.name}'


class Connection(PartialConnection):
    """Represents a Discord profile connection.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: x == y

            Checks if two connections are equal.

        .. describe:: x != y

            Checks if two connections are not equal.

        .. describe:: hash(x)

            Return the connection's hash.

        .. describe:: str(x)

            Returns the connection's name.

    Attributes
    ----------
    revoked: :class:`bool`
        Whether the connection is revoked.
    friend_sync: :class:`bool`
        Whether friends are synced over the connection.
    show_activity: :class:`bool`
        Whether activities from this connection will be shown in presences.
    two_way_link: :class:`bool`
        Whether the connection is authorized both ways (i.e. it's both a connection and an authorization).
    metadata_visible: :class:`bool`
        Whether the connection's metadata is visible.
    metadata: Optional[:class:`Metadata`]
        Various metadata about the connection.

        The contents of this are always subject to change.
    access_token: Optional[:class:`str`]
        The OAuth2 access token for the account, if applicable.
    integrations: List[:class:`Integration`]
        The integrations attached to the connection.
    """

    __slots__ = (
        '_state',
        'revoked',
        'friend_sync',
        'show_activity',
        'two_way_link',
        'metadata_visible',
        'access_token',
        'integrations',
    )

    def __init__(self, *, data: ConnectionPayload, state: ConnectionState):
        self._update(data)
        self._state = state
        self.access_token: Optional[str] = None

    def _update(self, data: ConnectionPayload):
        super()._update(data)
        self.revoked: bool = data.get('revoked', False)
        self.visible: bool = bool(data.get('visibility', False))
        self.friend_sync: bool = data.get('friend_sync', False)
        self.show_activity: bool = data.get('show_activity', True)
        self.two_way_link: bool = data.get('two_way_link', False)
        self.metadata_visible: bool = bool(data.get('metadata_visibility', False))

        # Only sometimes in the payload
        try:
            self.access_token: Optional[str] = data['access_token']
        except KeyError:
            pass

        self.integrations: List[Integration] = [
            Integration(data=i, guild=self._resolve_guild(i)) for i in data.get('integrations') or []
        ]

    def _resolve_guild(self, data: IntegrationPayload) -> Guild:
        from .guild import Guild

        state = self._state
        guild_data = data.get('guild')
        if not guild_data:
            return None  # type: ignore

        guild_id = int(guild_data['id'])
        guild = state._get_guild(guild_id)
        if guild is None:
            guild = Guild(data=guild_data, state=state)
        return guild

    async def edit(
        self,
        *,
        name: str = MISSING,
        visible: bool = MISSING,
        friend_sync: bool = MISSING,
        show_activity: bool = MISSING,
        metadata_visible: bool = MISSING,
    ) -> Connection:
        """|coro|

        Edit the connection.

        All parameters are optional.

        Parameters
        ----------
        name: :class:`str`
            The new name of the connection. Only editable for certain connection types.
        visible: :class:`bool`
            Whether the connection is visible on your profile.
        show_activity: :class:`bool`
            Whether activities from this connection will be shown in presences.
        friend_sync: :class:`bool`
            Whether friends are synced over the connection.
        metadata_visible: :class:`bool`
            Whether the connection's metadata is visible.

        Raises
        ------
        HTTPException
            Editing the connection failed.

        Returns
        -------
        :class:`Connection`
            The edited connection.
        """
        payload = {}
        if name is not MISSING:
            payload['name'] = name
        if visible is not MISSING:
            payload['visibility'] = visible
        if show_activity is not MISSING:
            payload['show_activity'] = show_activity
        if friend_sync is not MISSING:
            payload['friend_sync'] = friend_sync
        if metadata_visible is not MISSING:
            payload['metadata_visibility'] = metadata_visible
        data = await self._state.http.edit_connection(self.type.value, self.id, **payload)
        return Connection(data=data, state=self._state)

    async def refresh(self) -> None:
        """|coro|

        Refreshes the connection. This updates the connection's :attr:`metadata`.

        Raises
        ------
        HTTPException
            Refreshing the connection failed.
        """
        await self._state.http.refresh_connection(self.type.value, self.id)

    async def delete(self) -> None:
        """|coro|

        Removes the connection.

        Raises
        ------
        HTTPException
            Deleting the connection failed.
        """
        await self._state.http.delete_connection(self.type.value, self.id)

    async def fetch_access_token(self) -> str:
        """|coro|

        Retrieves a new access token for the connection.
        Only applicable for connections of type:attr:`ConnectionType.twitch`,
        :attr:`ConnectionType.youtube`, and :attr:`ConnectionType.spotify`.

        Raises
        ------
        HTTPException
            Retrieving the access token failed.

        Returns
        -------
        :class:`str`
            The new access token.
        """
        data = await self._state.http.get_connection_token(self.type.value, self.id)
        return data['access_token']



"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations
from typing import Literal, TYPE_CHECKING, List, Optional, Tuple, Type, Union
import unicodedata

from .mixins import Hashable
from .asset import Asset, AssetMixin
from .utils import cached_slot_property, find, snowflake_time, get, MISSING, _get_as_snowflake
from .errors import InvalidData
from .enums import StickerType, StickerFormatType, try_enum

__all__ = (
    'StickerPack',
    'StickerItem',
    'Sticker',
    'StandardSticker',
    'GuildSticker',
)

if TYPE_CHECKING:
    import datetime
    from .state import ConnectionState
    from .user import User
    from .guild import Guild
    from .types.sticker import (
        StickerPack as StickerPackPayload,
        StickerItem as StickerItemPayload,
        Sticker as StickerPayload,
        StandardSticker as StandardStickerPayload,
        GuildSticker as GuildStickerPayload,
    )


class StickerPack(Hashable):
    """Represents a sticker pack.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: str(x)

            Returns the name of the sticker pack.

        .. describe:: x == y

           Checks if the sticker pack is equal to another sticker pack.

        .. describe:: x != y

           Checks if the sticker pack is not equal to another sticker pack.

    Attributes
    -----------
    name: :class:`str`
        The name of the sticker pack.
    description: :class:`str`
        The description of the sticker pack.
    id: :class:`int`
        The id of the sticker pack.
    stickers: List[:class:`StandardSticker`]
        The stickers of this sticker pack.
    sku_id: :class:`int`
        The SKU ID of the sticker pack.
    cover_sticker_id: Optional[:class:`int`]
         The ID of the sticker used for the cover of the sticker pack.
    cover_sticker: Optional[:class:`StandardSticker`]
        The sticker used for the cover of the sticker pack.
    """

    __slots__ = (
        '_state',
        'id',
        'stickers',
        'name',
        'sku_id',
        'cover_sticker_id',
        'cover_sticker',
        'description',
        '_banner',
    )

    def __init__(self, *, state: ConnectionState, data: StickerPackPayload) -> None:
        self._state: ConnectionState = state
        self._from_data(data)

    def _from_data(self, data: StickerPackPayload) -> None:
        self.id: int = int(data['id'])
        stickers = data['stickers']
        self.stickers: List[StandardSticker] = [StandardSticker(state=self._state, data=sticker) for sticker in stickers]
        self.name: str = data['name']
        self.sku_id: int = int(data['sku_id'])
        self.cover_sticker_id: Optional[int] = _get_as_snowflake(data, 'cover_sticker_id')
        self.cover_sticker: Optional[StandardSticker] = get(self.stickers, id=self.cover_sticker_id)
        self.description: str = data['description']
        self._banner: Optional[int] = _get_as_snowflake(data, 'banner_asset_id')

    @property
    def banner(self) -> Optional[Asset]:
        """:class:`Asset`: The banner asset of the sticker pack."""
        return self._banner and Asset._from_sticker_banner(self._state, self._banner)  # type: ignore

    def __repr__(self) -> str:
        return f'<StickerPack id={self.id} name={self.name!r} description={self.description!r}>'

    def __str__(self) -> str:
        return self.name


class _StickerTag(Hashable, AssetMixin):
    __slots__ = ()

    id: int
    format: StickerFormatType

    async def read(self) -> bytes:
        """|coro|

        Retrieves the content of this sticker as a :class:`bytes` object.

        .. note::

            Stickers that use the :attr:`StickerFormatType.lottie` format cannot be read.

        Raises
        ------
        HTTPException
            Downloading the asset failed.
        NotFound
            The asset was deleted.
        TypeError
            The sticker is a lottie type.

        Returns
        -------
        :class:`bytes`
            The content of the asset.
        """
        if self.format is StickerFormatType.lottie:
            raise TypeError('Cannot read stickers of format "lottie"')
        return await super().read()


class StickerItem(_StickerTag):
    """Represents a sticker item.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: str(x)

            Returns the name of the sticker item.

        .. describe:: x == y

           Checks if the sticker item is equal to another sticker item.

        .. describe:: x != y

           Checks if the sticker item is not equal to another sticker item.

    Attributes
    -----------
    name: :class:`str`
        The sticker's name.
    id: :class:`int`
        The id of the sticker.
    format: :class:`StickerFormatType`
        The format for the sticker's image.
    url: :class:`str`
        The URL for the sticker's image.
    """

    __slots__ = ('_state', 'name', 'id', 'format', 'url')

    def __init__(self, *, state: ConnectionState, data: StickerItemPayload):
        self._state: ConnectionState = state
        self.name: str = data['name']
        self.id: int = int(data['id'])
        self.format: StickerFormatType = try_enum(StickerFormatType, data['format_type'])
        self.url: str = f'{Asset.BASE}/stickers/{self.id}.{self.format.file_extension}'

    def __repr__(self) -> str:
        return f'<StickerItem id={self.id} name={self.name!r} format={self.format}>'

    def __str__(self) -> str:
        return self.name

    async def fetch(self) -> Union[Sticker, StandardSticker, GuildSticker]:
        """|coro|

        Attempts to retrieve the full sticker data of the sticker item.

        Raises
        --------
        HTTPException
            Retrieving the sticker failed.

        Returns
        --------
        Union[:class:`StandardSticker`, :class:`GuildSticker`]
            The retrieved sticker.
        """
        data = await self._state.http.get_sticker(self.id)
        cls, _ = _sticker_factory(data['type'])
        return cls(state=self._state, data=data)

    async def fetch_guild(self):
        """|coro|

        Retrieves the guild this sticker belongs to.

        Raises
        ------
        NotFound
            The guild this sticker belongs to is not public.
        HTTPException
            An error occurred while fetching the guild.

        Returns
        -------
        :class:`Guild`
            The guild this emoji belongs to.
        """
        from .guild import Guild  # Circular import

        state = self._state
        data = await state.http.get_sticker_guild(self.id)
        return Guild(state=state, data=data)


class Sticker(_StickerTag):
    """Represents a sticker.

    .. versionadded:: 1.6

    .. container:: operations

        .. describe:: str(x)

            Returns the name of the sticker.

        .. describe:: x == y

           Checks if the sticker is equal to another sticker.

        .. describe:: x != y

           Checks if the sticker is not equal to another sticker.

    Attributes
    ----------
    name: :class:`str`
        The sticker's name.
    id: :class:`int`
        The id of the sticker.
    description: :class:`str`
        The description of the sticker.
    pack_id: :class:`int`
        The id of the sticker's pack.
    format: :class:`StickerFormatType`
        The format for the sticker's image.
    url: :class:`str`
        The URL for the sticker's image.
    """

    __slots__ = ('_state', 'id', 'name', 'description', 'format', 'url')

    def __init__(self, *, state: ConnectionState, data: StickerPayload) -> None:
        self._state: ConnectionState = state
        self._from_data(data)

    def _from_data(self, data: StickerPayload) -> None:
        self.id: int = int(data['id'])
        self.name: str = data['name']
        self.description: str = data['description']
        self.format: StickerFormatType = try_enum(StickerFormatType, data['format_type'])
        self.url: str = f'{Asset.BASE}/stickers/{self.id}.{self.format.file_extension}'

    def __repr__(self) -> str:
        return f'<Sticker id={self.id} name={self.name!r}>'

    def __str__(self) -> str:
        return self.name

    @property
    def created_at(self) -> datetime.datetime:
        """:class:`datetime.datetime`: Returns the sticker's creation time in UTC."""
        return snowflake_time(self.id)


class StandardSticker(Sticker):
    """Represents a sticker that is found in a standard sticker pack.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: str(x)

            Returns the name of the sticker.

        .. describe:: x == y

           Checks if the sticker is equal to another sticker.

        .. describe:: x != y

           Checks if the sticker is not equal to another sticker.

    Attributes
    ----------
    name: :class:`str`
        The sticker's name.
    id: :class:`int`
        The id of the sticker.
    description: :class:`str`
        The description of the sticker.
    pack_id: :class:`int`
        The id of the sticker's pack.
    format: :class:`StickerFormatType`
        The format for the sticker's image.
    tags: List[:class:`str`]
        A list of tags for the sticker.
    sort_value: :class:`int`
        The sticker's sort order within its pack.
    """

    __slots__ = ('sort_value', 'pack_id', 'type', 'tags')

    def _from_data(self, data: StandardStickerPayload) -> None:
        super()._from_data(data)
        self.sort_value: int = data['sort_value']
        self.pack_id: int = int(data['pack_id'])
        self.type: StickerType = StickerType.standard

        try:
            self.tags: List[str] = [tag.strip() for tag in data['tags'].split(',')]
        except KeyError:
            self.tags = []

    def __repr__(self) -> str:
        return f'<StandardSticker id={self.id} name={self.name!r} pack_id={self.pack_id}>'

    async def pack(self) -> StickerPack:
        """|coro|

        Retrieves the sticker pack that this sticker belongs to.

        Raises
        --------
        InvalidData
            The corresponding sticker pack was not found.
        HTTPException
            Retrieving the sticker pack failed.

        Returns
        --------
        :class:`StickerPack`
            The retrieved sticker pack.
        """
        data = await self._state.http.list_premium_sticker_packs()
        packs = data['sticker_packs']
        pack = find(lambda d: int(d['id']) == self.pack_id, packs)

        if pack:
            return StickerPack(state=self._state, data=pack)
        raise InvalidData(f'Could not find corresponding sticker pack for {self!r}')


class GuildSticker(Sticker):
    """Represents a sticker that belongs to a guild.

    .. versionadded:: 2.0

    .. container:: operations

        .. describe:: str(x)

            Returns the name of the sticker.

        .. describe:: x == y

           Checks if the sticker is equal to another sticker.

        .. describe:: x != y

           Checks if the sticker is not equal to another sticker.

    Attributes
    ----------
    name: :class:`str`
        The sticker's name.
    id: :class:`int`
        The id of the sticker.
    description: :class:`str`
        The description of the sticker.
    format: :class:`StickerFormatType`
        The format for the sticker's image.
    available: :class:`bool`
        Whether this sticker is available for use.
    guild_id: :class:`int`
        The ID of the guild that this sticker is from.
    user: Optional[:class:`User`]
        The user that created this sticker. This can only be retrieved using :meth:`Guild.fetch_sticker` and
        having :attr:`~Permissions.manage_emojis_and_stickers`.
    emoji: :class:`str`
        The name of a unicode emoji that represents this sticker.
    """

    __slots__ = ('available', 'guild_id', 'user', 'emoji', 'type', '_cs_guild')

    def _from_data(self, data: GuildStickerPayload) -> None:
        super()._from_data(data)
        self.available: bool = data.get('available', True)
        self.guild_id: int = int(data['guild_id'])
        user = data.get('user')
        self.user: Optional[User] = self._state.store_user(user) if user else None
        self.emoji: str = data['tags']
        self.type: StickerType = StickerType.guild

    def __repr__(self) -> str:
        return f'<GuildSticker name={self.name!r} id={self.id} guild_id={self.guild_id} user={self.user!r}>'

    @cached_slot_property('_cs_guild')
    def guild(self) -> Optional[Guild]:
        """Optional[:class:`Guild`]: The guild that this sticker is from.
        Could be ``None`` if the bot is not in the guild.

        .. versionadded:: 2.0
        """
        return self._state._get_guild(self.guild_id)

    async def edit(
        self,
        *,
        name: str = MISSING,
        description: str = MISSING,
        emoji: str = MISSING,
        reason: Optional[str] = None,
    ) -> GuildSticker:
        """|coro|

        Edits a :class:`GuildSticker` for the guild.

        Parameters
        -----------
        name: :class:`str`
            The sticker's new name. Must be at least 2 characters.
        description: Optional[:class:`str`]
            The sticker's new description. Can be ``None``.
        emoji: :class:`str`
            The name of a unicode emoji that represents the sticker's expression.
        reason: :class:`str`
            The reason for editing this sticker. Shows up on the audit log.

        Raises
        -------
        Forbidden
            You are not allowed to edit stickers.
        HTTPException
            An error occurred editing the sticker.

        Returns
        --------
        :class:`GuildSticker`
            The newly modified sticker.
        """
        payload = {}

        if name is not MISSING:
            payload['name'] = name

        if description is not MISSING:
            payload['description'] = description

        if emoji is not MISSING:
            try:
                emoji = unicodedata.name(emoji)
            except TypeError:
                pass
            else:
                emoji = emoji.replace(' ', '_')

            payload['tags'] = emoji

        data = await self._state.http.modify_guild_sticker(self.guild_id, self.id, payload, reason)
        return GuildSticker(state=self._state, data=data)

    async def delete(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        Deletes the custom :class:`Sticker` from the guild.

        You must have :attr:`~Permissions.manage_emojis_and_stickers` to do this.

        Parameters
        -----------
        reason: Optional[:class:`str`]
            The reason for deleting this sticker. Shows up on the audit log.

        Raises
        -------
        Forbidden
            You are not allowed to delete stickers.
        HTTPException
            An error occurred deleting the sticker.
        """
        await self._state.http.delete_guild_sticker(self.guild_id, self.id, reason)

    async def fetch_guild(self):
        """|coro|

        Retrieves the guild this sticker belongs to.

        Raises
        ------
        NotFound
            The guild this sticker belongs to is not public.
        HTTPException
            An error occurred while fetching the guild.

        Returns
        -------
        :class:`Guild`
            The guild this emoji belongs to.
        """
        from .guild import Guild  # Circular import

        state = self._state
        data = await state.http.get_sticker_guild(self.id)
        return Guild(state=state, data=data)


def _sticker_factory(sticker_type: Literal[1, 2]) -> Tuple[Type[Union[StandardSticker, GuildSticker, Sticker]], StickerType]:
    value = try_enum(StickerType, sticker_type)
    if value == StickerType.standard:
        return StandardSticker, value
    elif value == StickerType.guild:
        return GuildSticker, value
    else:
        return Sticker, value
